#include <stdio.h>
#include <stdlib.h>
#include <strings.h>

#include "fmt_chunk.h"

#define BUFFER_SIZE 131072
#define DIRTY_ITERATIONS 50
#define DIRTY_COUNT 20000
#define DIRTY_INCREMENT 2000

#define DIRTY_THRESHOLD 20000
#define DIRTY_DECREMENT 1000

#define WAV_DATA_LEN 32768
/* #define USE_FPU 1 */

struct wave_t
{
/*  long start_of_wave; */
  int peak;
  int wave_len;
  int threshold;
  int skip_bytes;
  short wave_data[WAV_DATA_LEN];
  int pass;
};

/*

This module will assume stereo data now.

*/

int read_chars(FILE *in, char *s, int n)
{
int t,ch;

  for (t=0; t<n; t++)
  {
    ch=getc(in);
    if (ch==EOF) return -1;
    s[t]=ch;
  }
  s[t]=0;

  return 0;
}

int write_chars(FILE *out, char *s)
{
int t;

  t=0;
  while(s[t]!=0 && t<255)
  {
    putc(s[t++],out);
  }

  return 0;
}

int write_int(FILE *out, int n)
{
  putc((n&255),out);
  putc(((n>>8)&255),out);
  putc(((n>>16)&255),out);
  putc(((n>>24)&255),out);

  return 0;
}

int write_word(FILE *out, int n)
{
  putc((n&255),out);
  putc(((n>>8)&255),out);

  return 0;
}

unsigned int read_int(FILE *in)
{
unsigned int t;

  t=getc(in);
  t=(getc(in)<<8)+t;
  t=(getc(in)<<16)+t;
  t=(getc(in)<<24)+t;

  return t;
}

unsigned short int read_word(FILE *in)
{
unsigned short int t;

  t=getc(in);
  t=(getc(in)<<8)+t;

  return t;
}

int parse_header(FILE *in, FILE *out)
{
int length;
char riff_type[5];

  length=read_int(in);
  read_chars(in,riff_type,4);

  write_chars(out,"RIFF");
  write_int(out,length);
  write_chars(out,riff_type);

  printf("RIFF Header\n");
  printf("----------------------------\n");
  printf("          Length: %d\n",length);
  printf("            Type: %s\n",riff_type);
  printf("----------------------------\n");

  return 0;
}

int parse_fmt(FILE *in, FILE *out, struct fmt_chunk_t *fmt_chunk)
{
int length;

  length=read_int(in);
  fmt_chunk->format_type=read_word(in);
  fmt_chunk->channel_numbers=read_word(in);
  fmt_chunk->sample_rate=read_int(in);
  fmt_chunk->bytes_per_second=read_int(in);
  fmt_chunk->bytes_per_sample=read_word(in);
  fmt_chunk->bits_per_sample=read_word(in);

  write_chars(out,"fmt ");
  write_int(out,length);
  write_word(out,fmt_chunk->format_type);
  write_word(out,fmt_chunk->channel_numbers);
  write_int(out,fmt_chunk->sample_rate);
  write_int(out,fmt_chunk->bytes_per_second);
  write_word(out,fmt_chunk->bytes_per_sample);
  write_word(out,fmt_chunk->bits_per_sample);

  printf("FMT Chunk\n");
  printf("----------------------------\n");
  printf("          Length: %d\n",length);
  printf("     Format Type: ");
  if (fmt_chunk->format_type==0)
  { printf("Mono\n"); }
    else
  if (fmt_chunk->format_type==1)
  { printf("Stereo\n"); }
    else
  { printf("unknown\n"); }

  printf(" Channel Numbers: %d\n",fmt_chunk->channel_numbers);
  printf("     Sample Rate: %d\n",fmt_chunk->sample_rate);
  printf("Bytes Per Second: %d\n",fmt_chunk->bytes_per_second);
  printf("Bytes Per Sample: ");
  if (fmt_chunk->bytes_per_sample==1)
  { printf("8 bit mono (%d)\n",fmt_chunk->bytes_per_sample); }
    else
  if (fmt_chunk->bytes_per_sample==2)
  { printf("8 bit stereo or 16 bit mono (%d)\n",fmt_chunk->bytes_per_sample); }
    else
  if (fmt_chunk->bytes_per_sample==4)
  { printf("16 bit stereo (%d)\n",fmt_chunk->bytes_per_sample); }

  printf(" Bits Per Sample: %d\n",fmt_chunk->bits_per_sample);
  printf("----------------------------\n");

  return 0;
}

#ifdef OLD_NORMALIZE
int parse_data_old(FILE *in, struct fmt_chunk_t *fmt_chunk, int dirty)
{
short int buffer[BUFFER_SIZE];
unsigned int spikes1[DIRTY_COUNT*2+10];
unsigned int spikes2[DIRTY_COUNT*2+10];
unsigned int spikes1_peak[DIRTY_COUNT+10];
unsigned int spikes2_peak[DIRTY_COUNT+10];
unsigned int delta,spike_ptr;
char hill;
int *spike_curr_ptr,*spike_last_ptr;
int *spike_peak_curr_ptr,*spike_peak_last_ptr;
int *temp;
long marker;
int length,deepest;
int t,ptr,k,dirty_offset,dirty_count,iterations;
int last_dirty_count,last_dirty_offset,dirty_flag;
short int r;
int p;
int r1;
float ratio,tempratio=0;

  deepest=0;
  iterations=0;
  dirty_offset=DIRTY_INCREMENT;
  last_dirty_offset=0;
  last_dirty_count=0;

  length=read_int(in);
  marker=ftell(in);

  spikes1[0]=-1;
  spikes2[0]=-1;
  spike_curr_ptr=spikes1;
  spike_last_ptr=spikes2;
  spike_peak_curr_ptr=spikes1_peak;
  spike_peak_last_ptr=spikes2_peak;

  printf("DATA chunk\n");
  printf("----------------------------\n");
  printf("          Length: %d\n",length);

  printf("Scanning for biggest/smallest peak\n");

  if (fmt_chunk->bits_per_sample==16)
  {
    for (t=0; t<length/2; t++)
    {
      r=getc(in)+(getc(in)<<8);
      p=abs(r);
      if (p>deepest) deepest=p;
    }
  }

  printf("Clean Deepest: %d\n",deepest);

  if (deepest>32760 && dirty==0)
  {
    printf("This song is already normalized.\n");
    return -1;
  }

  if (dirty==1)
  {
    while(1)
    {
      fseek(in,marker,SEEK_SET);

      dirty_count=0;
      spike_ptr=0;
      dirty_flag=0;
      delta=0;
      hill=1;

      /* for (t=0; t<DIRTY_COUNT; t++) spike_curr_ptr[t]=0; */

      for (t=0; t<length/2; t++)
      {
        r=getc(in)+(getc(in)<<8);
        if (r<0 && hill==1)
        {
          hill=0;

          /* if (dirty_flag==-1) */
          if (dirty_flag!=0)
          {
            spike_curr_ptr[spike_ptr++]=delta;
            spike_curr_ptr[spike_ptr++]=t;
            dirty_flag=0;
          }

          delta=t;
        }
          else
        if (r>=0 && hill==0)
        {
          hill=1;

          /* if (dirty_flag==-1) */
          if (dirty_flag!=0)
          {
            spike_curr_ptr[spike_ptr]=delta;
            spike_ptr++;
            spike_curr_ptr[spike_ptr]=t;
            spike_ptr++;
            dirty_flag=0;
          }

          delta=t;
        }

        p=abs(r);
        if (p>deepest-dirty_offset && (dirty_flag==0 || dirty_flag==-1))
        {
          dirty_count++;

          if (dirty_flag==0)
          { spike_peak_curr_ptr[spike_ptr>>1]=p; }
            else
          if (p>spike_peak_curr_ptr[spike_ptr>>1]) 
          { spike_peak_curr_ptr[spike_ptr>>1]=p; }

          dirty_flag=1;
        }
          else
        if (p<deepest-dirty_offset && dirty_flag==1)
        {
          dirty_flag=-1;
        }
          else
        if (p>deepest-dirty_offset && dirty_flag==1)
        {
          if (p>spike_peak_curr_ptr[spike_ptr>>1]) 
          { spike_peak_curr_ptr[spike_ptr>>1]=p; }
        }

        if (dirty_count>DIRTY_COUNT) break;
      }

      printf("Dirty peak counts: %d\n",dirty_count);

      spike_curr_ptr[spike_ptr]=-1;
      spike_curr_ptr[spike_ptr+1]=-1;

      temp=spike_peak_curr_ptr;
      spike_peak_curr_ptr=spike_peak_last_ptr;
      spike_peak_last_ptr=temp;

      temp=spike_curr_ptr;
      spike_curr_ptr=spike_last_ptr;
      spike_last_ptr=temp;

      if (dirty_count>DIRTY_COUNT) break;

      last_dirty_count=dirty_count;
      last_dirty_offset=dirty_offset;
      dirty_offset=dirty_offset+DIRTY_INCREMENT;

      iterations++;
      if (iterations>=DIRTY_ITERATIONS) break;
    }

    deepest=deepest-last_dirty_offset;
    printf("Dirty deepest: %d\n",deepest-last_dirty_offset);

    if (deepest>32760)
    {
      printf("This song is already normalized.\n");
      return -1;
    }
  }

  fseek(in,marker,SEEK_SET);

  spike_ptr=0;
  hill=0;

  ratio=32767/(float)deepest;
  printf("Ratio: %.5f\n",ratio);
  printf("Rewriting wave\n");

  /* Write Data Chunk */

#ifdef DEBUG
  for (t=0; t<42; t++)
  {
    if (t%2==0) printf("(");
    printf("%d ",spike_curr_ptr[t]);
    if (t%2==1) printf(")");
  }
  printf("\n");
#endif

  if (fmt_chunk->bits_per_sample==16)
  {
    ptr=0;

    for (t=0; t<length/2; t++)
    {
      r=getc(in)+(getc(in)<<8);

      if (spike_curr_ptr[spike_ptr]==t)
      {
        if (hill==1)
        {
          hill=0;
#ifdef DEBUG
          printf("Hill end %d\n\n",t);
#endif
          spike_ptr++;

          if (spike_curr_ptr[spike_ptr]==t)
          {
            hill=1;
            tempratio=32767/(float)spike_peak_curr_ptr[spike_ptr>>1];
            spike_ptr++;
#ifdef DEBUG
            printf("Hill start %d new ratio %.5f with a peak %d\n",t,tempratio,spike_peak_curr_ptr[spike_ptr>>1]);
#endif
          }
        }
          else
        {
#ifdef DEBUG
if (spike_ptr%2!=0) printf("Warning: Spike's out of sync\n");
#endif
          hill=1;
          tempratio=32767/(float)spike_peak_curr_ptr[spike_ptr>>1];
          spike_ptr++;
#ifdef DEBUG
          printf("Hill start %d new ratio %.5f with a peak %d\n",t,tempratio,spike_peak_curr_ptr[spike_ptr>>1]);
#endif
        }
      }

      if (hill==0)
      { r1=(int)((float)r*ratio); }
        else
      { r1=(int)((float)r*tempratio); }

      if (r1>32767)
      {
        printf("Clip warning at: %d   volume: %d  r=%d  ratio=%.5f  hill=%d\n",t,r1,r,ratio,hill);
        r1=32767;
      }
        else
      if (r1<-32767)
      {
        printf("Clip warning at: %d   volume: %d  r=%d  ratio=%.5f  hill=%d\n",t,r1,r,ratio,hill);
        r1=-32767;
      }

      r=r1;

      buffer[ptr++]=r;
      if (ptr>=BUFFER_SIZE)
      {
        fseek(in,-(ptr*2), SEEK_CUR);
        for (k=0; k<ptr; k++)
        { write_word(in,buffer[k]); }
        ptr=0;
      }
/*
      fseek(in,-2, SEEK_CUR);
      write_word(in,r);
*/
    }

#ifdef DEBUG
    printf("next hill starts at: %d\n",spike_curr_ptr[spike_ptr]);
#endif

    if (ptr!=0)
    {
      fseek(in,-(ptr*2), SEEK_CUR);
      for (k=0; k<ptr; k++)
      { write_word(in,buffer[k]); }
    }
  }
    else
  if (fmt_chunk->bits_per_sample==8)
  {
    printf("8 bit wav's not supported\n");
  }

  return 0;
}
#endif

int lower_threshold(FILE *in, long marker, int max_dirty_peaks, int threshold, int skip_bytes, int samples)
{
int t;
int dirty_peaks;
int new_threshold;
short int curr_sample;
int hill;
int dirty_flag;

  new_threshold=threshold;

  while(1)
  {
    new_threshold=new_threshold-DIRTY_DECREMENT;
    dirty_peaks=0;
    dirty_flag=0;
    hill=1;

    fseek(in,marker,SEEK_SET);

    for (t=0; t<samples; t++)
    {
      curr_sample=getc(in)+(getc(in)<<8);
      if (skip_bytes==2) { getc(in); getc(in); }

      if (hill==1 && curr_sample<0)
      { dirty_flag=0; hill=-1; }
        else
      if (hill==-1 && curr_sample>0)
      { dirty_flag=0; hill=1; }

      if (abs(curr_sample)>new_threshold && dirty_flag==0)
      {
        dirty_flag=1;
        dirty_peaks++;
      }
    }

    if (dirty_peaks>max_dirty_peaks) break;

    threshold=new_threshold;

    if (threshold<3000) break;
  }

  return threshold;
}

int write_wave(FILE *out, struct wave_t *wave)
{
/* long marker; */
int t;
short curr_sample;
#ifdef USE_FPU
float ratio;
#else
int int_ratio;
#endif

  /* marker=ftell(in); */

  /* fseek(in,wave->start_of_wave,SEEK_SET); */

  if (wave->peak<=wave->threshold)
  {
#ifdef USE_FPU
    ratio=(32767/(float)wave->threshold);
#else
    int_ratio=wave->threshold;
#endif
  }
    else
  {
#ifdef USE_FPU
    ratio=(32767/(float)wave->peak);
#else
    int_ratio=wave->peak;
#endif
  }

  for (t=0; t<wave->wave_len; t++)
  {
#ifdef USE_FPU
    curr_sample=(short)((float)wave->wave_data[t]*ratio);
#else
    curr_sample=(short)((wave->wave_data[t]*32767)/int_ratio);
#endif

    putc(curr_sample&255,out);
    putc(curr_sample>>8,out);

    if (wave->skip_bytes==2)
    {
/*
      getc(out);
      getc(out);
*/
      if (wave->pass==1)
      /* { fseek(out,2,SEEK_CUR); } */
      { getc(out); getc(out); }
        else
      { putc(0,out); putc(0,out); }
/*
      putc(0,out);
      putc(0,out);
*/
    }
  }

  wave->peak=0;
  wave->wave_len=0;

  return 0;
}

int adjust_wav(FILE *in, FILE *out, long marker, int threshold, int skip_bytes, int samples, int pass)
{
short curr_sample;
float ratio;
struct wave_t wave;
int t;
int hill;
int total;

  ratio=32767/(float)threshold;
  printf("Ratio: %.5f  threshold=%d\n",ratio,threshold);

  fseek(in,marker,SEEK_SET);
  fseek(out,marker,SEEK_SET);

  hill=1;
  wave.wave_len=0;
  /* wave.start_of_wave=marker; */
  wave.peak=0;
  wave.skip_bytes=skip_bytes;
  wave.threshold=threshold;
  wave.pass=pass;

total=0;

  for (t=0; t<samples; t++)
  {
    curr_sample=getc(in)+(getc(in)<<8);
    if (skip_bytes==2) { getc(in); getc(in); }

/* printf("%d: %d\n",t,curr_sample); */

    if (hill==1 && curr_sample<0)
    {
total=total+wave.wave_len;
// printf("%d> %d\n",t,total);
      write_wave(out,&wave);
if (ftell(in)!=ftell(out)+4)
{
printf("ftell(in)=%d ftell(out)=%d\n",ftell(in),ftell(out)+4);
break;
}
      hill=-1;
/* printf("hill now %d\n",hill); */
    }
      else
    if (hill==-1 && curr_sample>=0)
    {
total=total+wave.wave_len;
// printf("%d> %d\n",t,total);
      write_wave(out,&wave);
if (ftell(in)!=ftell(out)+4)
{
printf("ftell(in)=%d ftell(out)=%d\n",ftell(in),ftell(out)+4);
break;
}
      hill=1;
/* printf("hill now %d\n",hill); */
    }

    wave.wave_data[wave.wave_len++]=curr_sample;
    if (abs(curr_sample)>wave.peak) wave.peak=abs(curr_sample);

    if (wave.wave_len>=WAV_DATA_LEN)
    {
      printf("Warning: wave_data overflow\n");
      write_wave(out,&wave);
      /* wave.start_of_wave=ftell(in); */
    }
  }

  if (wave.wave_len>0)
  { write_wave(out,&wave); }

  printf("ftell(in)=%d ftell(out)=%d\n",ftell(in),ftell(out));

  return 0;
}

int parse_data(FILE *in, FILE *out, struct fmt_chunk_t *fmt_chunk, int max_dirty_peaks)
{
int length,threshold;
long marker;
int samples;
int t,r;
short s;

  threshold=0;

  length=read_int(in);
  marker=ftell(in);

  printf("DATA chunk\n");
  printf("----------------------------\n");
  printf("          Length: %d\n",length);
  printf(" Max Dirty Peaks: %d\n",max_dirty_peaks);

  if (fmt_chunk->bits_per_sample!=16)
  {
    printf("Only 16 bit wav's are supported\n");
    return -1;
  }

  write_chars(out,"data");
  write_int(out,length);

  samples=length>>1;

  for (t=0; t<samples; t++)
  {
    s=abs(getc(in)+(getc(in)<<8));
    if (s>threshold) threshold=s;
  }

  printf("   Clean Deepest: %d\n",threshold);

  if (threshold>32760 && max_dirty_peaks==0)
  {
    printf("This song is already normalized.\n");
    return -1;
  }

  if (max_dirty_peaks>0)
  {
    if (fmt_chunk->channel_numbers==2)
    {
      printf("Analyzing channel 0\n");
      r=lower_threshold(in,marker,max_dirty_peaks,threshold,2,samples>>1);
      if (r<threshold) threshold=r;

      printf("Analyzing channel 1\n");
      r=lower_threshold(in,marker+2,max_dirty_peaks,threshold,2,samples>>1);
      if (r<threshold) threshold=r;
    }
      else
    {
      printf("Analyzing sound data\n");
      threshold=lower_threshold(in,marker,max_dirty_peaks,threshold,0,samples);
    }
  }

  fseek(in,marker,SEEK_SET);

  /* Write Data Chunk */

  if (fmt_chunk->channel_numbers==2)
  {
    printf("Adjusting channel 0\n");
    r=adjust_wav(in,out,marker,threshold,2,samples>>1,0);
    printf("Adjusting channel 1\n");
    r=adjust_wav(in,out,marker+2,threshold,2,samples>>1,1);
  }
    else
  {
    printf("Adjusting wave\n");
    r=adjust_wav(in,out,marker,threshold,0,samples,0);
  }

  return 0;
}

int normalize(char *inname, char *outname, int dirty)
{
FILE *in,*out;
char chunk_name[5];
struct fmt_chunk_t fmt_chunk;
long marker;

  in=fopen(inname,"rb");
  if (in==0)
  {
    printf("Couldn't open file for reading: %s\n",inname);
    return -1;
  }

  out=fopen(outname,"wb+");
  if (out==0)
  {
    printf("Couldn't open file for writing: %s\n",outname);
    fclose(in);
    return -1;
  }

  while(1)
  {
    if (read_chars(in,chunk_name,4)!=0) break;

    if (strcasecmp(chunk_name,"RIFF")==0)
    { parse_header(in,out); }
      else
    if (strcasecmp(chunk_name,"fmt ")==0)
    { parse_fmt(in,out,&fmt_chunk); }
      else
    if (strcasecmp(chunk_name,"data")==0)
    { if (parse_data(in,out,&fmt_chunk,dirty)==-1) break; }
      else
    {
      printf("Unknown chunk: '%s'\n",chunk_name);
      return -1;
    }
  }

  marker=ftell(out);
  fseek(out,4,SEEK_SET);
  write_int(out,marker);
  fseek(out,marker,SEEK_SET);

  fclose(out);
  fclose(in);

  return 0;
}

#ifdef STANDALONE
int main()
{
  normalize("/storage/test.wav","/storage/testnew.wav",DIRTY_THRESHOLD);
  return 0;
}
#endif

